# -*- coding: utf-8 -*-

class Node(str):
    pass

class Edge(object):
    def __init__(self, src, dest, weight = 1):
        """Assumes src and dest are nodes"""
        self._src = src
        self._dest = dest
        self._weight = weight
    def getSource(self):
        return self._src
    def getDestination(self):
        return self._dest
    def getWeight(self):
        return self._weight
    def __str__(self):
        return self.src + '->(' + self.getWeight() + ')'\
               + self.dest

class WeightedDigraph(object):
    """edges is a dict mapping each node to a list of
    its children and weight of edge"""
    def __init__(self, nodes):
        self._edges = {v: [] for v in nodes}
    def addNode(self, node):
        if node in self._edges:
            raise ValueError('Duplicate node')
        else:
            self._edges[node] = []
    def addEdge(self, edge):
        """edge is an Edge"""
        src = edge.getSource()
        dest = edge.getDestination()
        if not (src in self._edges and dest in self._edges):
            raise ValueError('Node not in graph')
        self._edges[src].append((dest, edge.getWeight()))
    def childrenOf(self, node):
        return [e[0] for e in self._edges[node]]
    def hasNode(self, node):
        return node in self._edges
    def getAllNodes(self):
        return(list(self._edges.keys()))   
    def __str__(self):
        vals = []
        for src in self._edges:
            entry = src + ': '
            for edge in self._edges[src]:
                entry += edge[0] + '(' + str(edge[1]) + '), '
            if entry[-2:] != ': ': #there was at least one edge
                vals.append(entry[:-2])
            else:
                vals.append(entry[:-1])
        vals.sort(key = lambda x: x.split(':')[0])
        result = ''
        for v in vals:
            result += v + '\n'
        return result[:-1] #omit final newline

def printPath(path):
    """Assumes path is a list of nodes"""
    if path == None:
        return 'None'
    result = ''
    for i in range(len(path)):
        result = result + path[i]
        if i != len(path) - 1:
            result = result + '->'
    return result 
    
def buildCityGraph():
    """Generate and return an example graph"""
    g = WeightedDigraph(('Boston', 'Providence', 'New York','Chicago',
                 'Denver', 'Phoenix', 'Los Angeles'))
    g.addEdge(Edge('Boston', 'Providence'))
    g.addEdge(Edge('Boston', 'New York'))
    g.addEdge(Edge('Providence', 'Boston'))
    g.addEdge(Edge('Providence', 'New York'))
    g.addEdge(Edge('New York', 'Chicago'))
    g.addEdge(Edge('Chicago', 'Denver'))
    g.addEdge(Edge('Chicago', 'Phoenix'))
    g.addEdge(Edge('Denver', 'Phoenix'))
    g.addEdge(Edge('Denver', 'New York'))
    g.addEdge(Edge('Los Angeles', 'Boston'))
    return g

g = buildCityGraph()
print('The city graph:')
print(g)

def bfs(graph, start, end, toPrint = False):
    """Assumes graph is a Digraph; start and end are nodes
       Returns a shortest path (number of hops) from start to end
       in graph"""
    visited = {} #no vertices visited thus far
    pathQueue = [[start]] #FIFO
    visited[start] = True
    while len(pathQueue) != 0:
        tmpPath = pathQueue.pop(0)
        if toPrint:
            print('Current BFS path:', printPath(tmpPath))
            print('Current queue:', pathQueue)
            print()
        lastNode = tmpPath[-1]
        if lastNode == end:
            return tmpPath
        for nextNode in graph.childrenOf(lastNode):
            if nextNode not in visited: #avoid visiting again
                newPath = tmpPath + [nextNode] #Note copy of list
                visited[nextNode] = True
                pathQueue.append(newPath)
    return None

g = buildCityGraph()
bfs(g, 'Boston', 'Phoenix', True)

def dfs(graph, start, end, toPrint = False):
    """Assumes graph is a Digraph; start and end are nodes
       Returns a shortest path from start to end in graph"""
    bestPath = None
    initPath = [start]
    pathQueue = [initPath] #LIFO
    while len(pathQueue) != 0:
        #Get and remove newest element in pathQueue
        tmpPath = pathQueue.pop(-1)
        if toPrint:
            print('Current DFS path:', printPath(tmpPath))
        lastNode = tmpPath[-1]
        if lastNode == end:
            if toPrint:
                print('Path found')
            if bestPath == None or len(tmpPath) < len(bestPath):
                bestPath = tmpPath
                continue
        if bestPath != None and len(tmpPath) >= len(bestPath):
            continue
        for nextNode in graph.childrenOf(lastNode):
            if nextNode not in tmpPath:
                newPath = tmpPath + [nextNode]
                pathQueue.append(newPath)
    return bestPath

#g = buildCityGraph()
#dfs(g, 'Boston', 'Phoenix', True)

def Dijkstra(graph, start, end, toPrint = False):
    """
    graph: an unweighted (all edges have weight 1) digraph
    start: a node in  graph
    end: a node in graph
    returns a list representing shortest path from start to end,
       and None if no path exists"""
    #Easily modified to deal with non-negative weighted edges

    # Mark all nodes unvisited and store them.
    # Set the distance to zero for our initial node 
    # and to infinity for other nodes.
    unvisited = graph.getAllNodes()
    distanceTo = {node: float('inf') for node in graph.getAllNodes()}
    distanceTo[start] = 0
    # Mark all nodes as not having found a predecessor node on path
    #from start
    predecessor = {node: None for node in graph.getAllNodes()}

    while unvisited:
        # Select the unvisited node with the smallest distance from 
        # start, it's current node now.
        current = min(unvisited, key=lambda node: distanceTo[node])
        if toPrint: #for pedagocical purposes
            print('\nValue of current:', current)
            print('Value of distanceTo:')
            for k in distanceTo:
                print('  ' + k + ':', distanceTo[k])
            print('Value of predecessor:')
            for k in predecessor:
                print('  ' + k + ':', predecessor[k])

        # Stop, if the smallest distance 
        # among the unvisited nodes is infinity.
        if distanceTo[current] == float('inf'):
            break

        # Find unvisited neighbors for the current node 
        # and calculate their distances from start through the
        # current node.
        for neighbour in graph.childrenOf(current):
            alternativePathDist = distanceTo[current] + 1 #hops as distance

            # Compare the newly calculated distance to the assigned. 
            # Save the smaller distance and update predecssor.
            if alternativePathDist < distanceTo[neighbour]:
                distanceTo[neighbour] = alternativePathDist
                predecessor[neighbour] = current

        # Remove the current node from the unvisited set.
        unvisited.remove(current)
            
    #Attempt to be build a path working backwards from end
    path = []
    current = end
    while predecessor[current] != None:
        path.insert(0, current)
        current = predecessor[current]
    if path != []:
        path.insert(0, current)
    else:
        return None
    return path

#g = buildCityGraph()
#path = Dijkstra(g, 'Boston', 'Chicago', toPrint = True)
#print(printPath(path))

def shortestPath(g, start, end, alg, toPrint = False):
    """Assumes graph is a Digraph; start and end are nodes
       Returns a shortest path from start to end in graph
       using algorithm alg."""
    sp = alg(g, start, end, toPrint)
    return sp

def printShortest(start, end, sp):
    if sp != None:
        print('Shortest path from', start, 'to',
              end, 'is', printPath(sp))
    else:
        print('There is no path from', start, 'to', end)


#G = buildCityGraph()
#print(G)
#for startName, endName in [('Chicago', 'Boston'), ('Boston', 'Phoenix')]:
#    sp = shortestPath(G, startName, endName, dfs, False)
#    printShortest(startName, endName, sp)
#    sp = shortestPath(G, startName, endName, bfs, False)
#    printShortest(startName, endName, sp)
#    sp = shortestPath(G, startName, endName, Dijkstra, False)
#    printShortest(startName, endName, sp)

