# -*- coding: utf-8 -*-

class Node(str):
    pass

class Edge(object):
    def __init__(self, src, dest, weight = 1):
        """Assumes src and dest are nodes"""
        self._src = src
        self._dest = dest
        self._weight = weight
    def getSource(self):
        return self._src
    def getDestination(self):
        return self._dest
    def getWeight(self):
        return self._weight
    def __str__(self):
        return self.src + '->(' + self.getWeight() + ')'\
               + self.dest

class WeightedDigraph(object):
    """edges is a dict mapping each node to a list of
    its children and weight of edge"""
    def __init__(self, nodes):
        self._edges = {v: [] for v in nodes}
    def addNode(self, node):
        if node in self._edges:
            raise ValueError('Duplicate node')
        else:
            self._edges[node] = []
    def addEdge(self, edge):
        """edge is an Edge"""
        src = edge.getSource()
        dest = edge.getDestination()
        if not (src in self._edges and dest in self._edges):
            raise ValueError('Node not in graph')
        self._edges[src].append((dest, edge.getWeight()))
    def childrenOf(self, node):
        return [e[0] for e in self._edges[node]]
    def hasNode(self, node):
        return node in self._edges
    def getAllNodes(self):
        return(list(self._edges.keys()))   
    def __str__(self):
        vals = []
        for src in self._edges:
            entry = src + ': '
            for edge in self._edges[src]:
                entry += edge[0] + '(' + str(edge[1]) + '), '
            if entry[-2:] != ': ': #there was at least one edge
                vals.append(entry[:-2])
            else:
                vals.append(entry[:-1])
        vals.sort(key = lambda x: x.split(':')[0])
        result = ''
        for v in vals:
            result += v + '\n'
        return result[:-1] #omit final newline

def printPath(path):
    """Assumes path is a list of nodes"""
    if path == None:
        return 'None'
    result = ''
    for i in range(len(path)):
        result = result + path[i]
        if i != len(path) - 1:
            result = result + '->'
    return result 
    

# the code below makes small changes to allow for weighted paths
# it also creates a new city graph with weighted edges
        
def Dijkstra_weighted(graph, start, end, toPrint = False):
    """
    graph: an unweighted (all edges have weight 1) digraph
    start: a node in  graph
    end: a node in graph
    returns a list representing shortest path from start to end,
       and None if no path exists"""
    #Easily modified to deal with non-negative weighted edges

    # Mark all nodes unvisited and store them.
    # Set the distance to zero for our initial node 
    # and to infinity for other nodes.
    unvisited = graph.getAllNodes()
    distanceTo = {node: float('inf') for node in graph.getAllNodes()}
    distanceTo[start] = 0
    # Mark all nodes as not having found a predecessor node on path
    #from start
    predecessor = {node: None for node in graph.getAllNodes()}

    while unvisited:
        # Select the unvisited node with the smallest distance from 
        # start, it's current node now.
        current = min(unvisited, key=lambda node: distanceTo[node])
        if toPrint: #for pedagocical purposes
            print('\nValue of current:', current)
            print('Value of distanceTo:')
            for k in distanceTo:
                print('  ' + k + ':', distanceTo[k])
            print('Value of predecessor:')
            for k in predecessor:
                print('  ' + k + ':', predecessor[k])

        # Stop, if the smallest distance 
        # among the unvisited nodes is infinity.
        if distanceTo[current] == float('inf'):
            break

        # Find unvisited neighbors for the current node 
        # and calculate their distances from start through the
        # current node.
        for edge in graph._edges[current]:
            alternativePathDist = distanceTo[current] + edge[1]
            neighbour = edge[0]
            # Compare the newly calculated distance to the assigned. 
            # Save the smaller distance and update predecssor.
            if alternativePathDist < distanceTo[neighbour]:
                distanceTo[neighbour] = alternativePathDist
                predecessor[neighbour] = current

        # Remove the current node from the unvisited set.
        unvisited.remove(current)
            
    #Attempt to be build a path working backwards from end
    path = []
    current = end
    while predecessor[current] != None:
        path.insert(0, current)
        current = predecessor[current]
    if path != []:
        path.insert(0, current)
    else:
        return None
    return path

def buildCityGraph_weighted():
    """Generate and return an example graph"""
    g = WeightedDigraph(('Boston', 'Providence', 'New York','Chicago',
                 'Denver', 'Phoenix', 'Los Angeles'))
    g.addEdge(Edge('Boston', 'Providence', 50))
    g.addEdge(Edge('Boston', 'New York', 190))
    g.addEdge(Edge('Providence', 'Boston', 50))
    g.addEdge(Edge('Providence', 'New York',180))
    g.addEdge(Edge('New York', 'Chicago', 790))
    g.addEdge(Edge('Chicago', 'Denver', 1000))
    g.addEdge(Edge('Chicago', 'Phoenix', 1750))
    g.addEdge(Edge('Denver', 'Phoenix', 820))
    g.addEdge(Edge('Denver', 'New York', 1780))
    g.addEdge(Edge('Los Angeles', 'Boston', 2990))
    return g

g = buildCityGraph_weighted()
print('The city graph:')
print(g)

path = Dijkstra_weighted(g, 'Boston', 'Phoenix', toPrint = True)
print(printPath(path))